/**
 * Copyright (c) 2010-2015, Grill Balázs, Istvan Rath and Daniel Varro
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * Contributors:
 * Grill Balázs - initial API and implementation
 */
package org.eclipse.viatra.query.testing.core;

import com.google.common.base.Objects;
import com.google.common.collect.Iterables;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import junit.framework.AssertionFailedError;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.ecore.resource.impl.ResourceSetImpl;
import org.eclipse.viatra.query.runtime.api.IPatternMatch;
import org.eclipse.viatra.query.runtime.api.IQueryGroup;
import org.eclipse.viatra.query.runtime.api.IQuerySpecification;
import org.eclipse.viatra.query.runtime.api.ViatraQueryMatcher;
import org.eclipse.viatra.query.testing.core.IMatchSetModelProvider;
import org.eclipse.viatra.query.testing.core.SnapshotHelper;
import org.eclipse.viatra.query.testing.core.internal.MatchSetRecordDiff;
import org.eclipse.viatra.query.testing.snapshot.MatchRecord;
import org.eclipse.viatra.query.testing.snapshot.MatchSetRecord;
import org.eclipse.xtext.xbase.lib.CollectionLiterals;
import org.eclipse.xtext.xbase.lib.Exceptions;
import org.eclipse.xtext.xbase.lib.Extension;
import org.eclipse.xtext.xbase.lib.Functions.Function1;
import org.eclipse.xtext.xbase.lib.IterableExtensions;
import org.eclipse.xtext.xbase.lib.Procedures.Procedure1;

/**
 * @author Grill Balázs
 */
@SuppressWarnings("all")
public class ViatraQueryTestCase {
  public final static String UNEXPECTED_MATCH = "Unexpected match";
  
  public final static String EXPECTED_NOT_FOUND = "Expected match not found";
  
  private final ResourceSet resourceSet;
  
  private final List<IMatchSetModelProvider> modelProviders;
  
  @Extension
  private SnapshotHelper _snapshotHelper = new SnapshotHelper();
  
  public ViatraQueryTestCase() {
    ResourceSetImpl _resourceSetImpl = new ResourceSetImpl();
    this.resourceSet = _resourceSetImpl;
    LinkedList<IMatchSetModelProvider> _linkedList = new LinkedList<IMatchSetModelProvider>();
    this.modelProviders = _linkedList;
  }
  
  public Resource loadModel(final URI uri) {
    return this.resourceSet.getResource(uri, true);
  }
  
  public void modifyModel(final URI uri, final Procedure1<? super Resource> operation) {
    Resource _resource = this.resourceSet.getResource(uri, false);
    operation.apply(_resource);
  }
  
  public boolean addMatchSetModelProvider(final IMatchSetModelProvider matchSetModelProvider) {
    return this.modelProviders.add(matchSetModelProvider);
  }
  
  public <Match extends IPatternMatch> void assertMatchSetsEqual(final IQuerySpecification<? extends ViatraQueryMatcher<Match>> querySpecification) {
    int _size = this.modelProviders.size();
    boolean _lessThan = (_size < 2);
    if (_lessThan) {
      throw new IllegalArgumentException("At least two model providers shall be set");
    }
    final IMatchSetModelProvider reference = IterableExtensions.<IMatchSetModelProvider>head(this.modelProviders);
    Iterable<IMatchSetModelProvider> _tail = IterableExtensions.<IMatchSetModelProvider>tail(this.modelProviders);
    final Procedure1<IMatchSetModelProvider> _function = new Procedure1<IMatchSetModelProvider>() {
      @Override
      public void apply(final IMatchSetModelProvider it) {
        final HashSet<String> diff = ViatraQueryTestCase.this.<Match>compareMatchSets(querySpecification, reference, it);
        boolean _isEmpty = diff.isEmpty();
        boolean _not = (!_isEmpty);
        if (_not) {
          String _string = diff.toString();
          throw new AssertionFailedError(_string);
        }
      }
    };
    IterableExtensions.<IMatchSetModelProvider>forEach(_tail, _function);
  }
  
  public void assertMatchSetsEqual(final IQueryGroup queryGroup) {
    Set<IQuerySpecification<?>> _specifications = queryGroup.getSpecifications();
    final Procedure1<IQuerySpecification<?>> _function = new Procedure1<IQuerySpecification<?>>() {
      @Override
      public void apply(final IQuerySpecification<?> it) {
        ViatraQueryTestCase.this.<IPatternMatch>assertMatchSetsEqual(((IQuerySpecification<? extends ViatraQueryMatcher<IPatternMatch>>) it));
      }
    };
    IterableExtensions.<IQuerySpecification<?>>forEach(_specifications, _function);
  }
  
  public <Match extends IPatternMatch> MatchSetRecordDiff getMatchSetDiff(final IQuerySpecification<? extends ViatraQueryMatcher<Match>> querySpecification, final IMatchSetModelProvider expectedProvider, final IMatchSetModelProvider actualProvider) {
    try {
      MatchSetRecordDiff _xblockexpression = null;
      {
        Match filter = null;
        MatchSetRecord expected = expectedProvider.<Match>getMatchSetRecord(this.resourceSet, querySpecification, filter);
        MatchRecord _filter = expected.getFilter();
        boolean _notEquals = (!Objects.equal(_filter, null));
        if (_notEquals) {
          MatchRecord _filter_1 = expected.getFilter();
          Match _createMatchForMatchRecord = this._snapshotHelper.<Match>createMatchForMatchRecord(querySpecification, _filter_1);
          filter = _createMatchForMatchRecord;
        }
        final MatchSetRecord actual = actualProvider.<Match>getMatchSetRecord(this.resourceSet, querySpecification, filter);
        MatchRecord _filter_2 = actual.getFilter();
        boolean _notEquals_1 = (!Objects.equal(_filter_2, null));
        if (_notEquals_1) {
          boolean _notEquals_2 = (!Objects.equal(filter, null));
          if (_notEquals_2) {
            throw new IllegalArgumentException(
              ((("Filter is provided by more than one sources: " + expectedProvider) + ", ") + actualProvider));
          } else {
            MatchRecord _filter_3 = actual.getFilter();
            Match _createMatchForMatchRecord_1 = this._snapshotHelper.<Match>createMatchForMatchRecord(querySpecification, _filter_3);
            filter = _createMatchForMatchRecord_1;
            MatchSetRecord _matchSetRecord = expectedProvider.<Match>getMatchSetRecord(this.resourceSet, querySpecification, filter);
            expected = _matchSetRecord;
          }
        }
        _xblockexpression = MatchSetRecordDiff.compute(expected, actual);
      }
      return _xblockexpression;
    } catch (Throwable _e) {
      throw Exceptions.sneakyThrow(_e);
    }
  }
  
  private <Match extends IPatternMatch> HashSet<String> compareMatchSets(final IQuerySpecification<? extends ViatraQueryMatcher<Match>> querySpecification, final IMatchSetModelProvider expectedProvider, final IMatchSetModelProvider actualProvider) {
    final MatchSetRecordDiff matchdiff = this.<Match>getMatchSetDiff(querySpecification, expectedProvider, actualProvider);
    final HashSet<String> diff = CollectionLiterals.<String>newHashSet();
    Set<MatchRecord> _additions = matchdiff.getAdditions();
    final Function1<MatchRecord, String> _function = new Function1<MatchRecord, String>() {
      @Override
      public String apply(final MatchRecord it) {
        String _prettyPrint = ViatraQueryTestCase.this._snapshotHelper.prettyPrint(it);
        String _plus = ((ViatraQueryTestCase.UNEXPECTED_MATCH + " (") + _prettyPrint);
        return (_plus + ")");
      }
    };
    Iterable<String> _map = IterableExtensions.<MatchRecord, String>map(_additions, _function);
    Iterables.<String>addAll(diff, _map);
    Set<MatchRecord> _removals = matchdiff.getRemovals();
    final Function1<MatchRecord, String> _function_1 = new Function1<MatchRecord, String>() {
      @Override
      public String apply(final MatchRecord it) {
        String _prettyPrint = ViatraQueryTestCase.this._snapshotHelper.prettyPrint(it);
        String _plus = ((ViatraQueryTestCase.EXPECTED_NOT_FOUND + " (") + _prettyPrint);
        return (_plus + ")");
      }
    };
    Iterable<String> _map_1 = IterableExtensions.<MatchRecord, String>map(_removals, _function_1);
    Iterables.<String>addAll(diff, _map_1);
    return diff;
  }
}
