/**
 * Copyright (c) 2010-2015, Grill Bal�zs, Istvan Rath and Daniel Varro
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * Contributors:
 * Grill Bal�zs - initial API and implementation
 */
package org.eclipse.viatra.query.testing.core.api;

import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import org.eclipse.emf.common.util.URI;
import org.eclipse.viatra.query.runtime.api.IPatternMatch;
import org.eclipse.viatra.query.runtime.api.IQueryGroup;
import org.eclipse.viatra.query.runtime.api.IQuerySpecification;
import org.eclipse.viatra.query.runtime.api.ViatraQueryMatcher;
import org.eclipse.viatra.query.runtime.extensibility.QuerySpecificationRegistry;
import org.eclipse.viatra.query.runtime.matchers.backend.IQueryBackend;
import org.eclipse.viatra.query.runtime.matchers.backend.QueryEvaluationHint;
import org.eclipse.viatra.query.testing.core.PatternBasedMatchSetModelProvider;
import org.eclipse.viatra.query.testing.core.SnapshotMatchSetModelProvider;
import org.eclipse.viatra.query.testing.core.ViatraQueryTestCase;
import org.eclipse.viatra.query.testing.core.XmiModelUtil;
import org.eclipse.xtext.xbase.lib.CollectionLiterals;
import org.eclipse.xtext.xbase.lib.IterableExtensions;
import org.eclipse.xtext.xbase.lib.Procedures.Procedure1;

/**
 * This class defines an API to easily construct test cases. The base conception is to provide
 * a set of match set sources (from a snapshot or from the execution of a pattern) and make assertions
 * on them being equal.
 */
@SuppressWarnings("all")
public class ViatraQueryTest {
  private final ViatraQueryTestCase testCase;
  
  /**
   * Test the specified query
   */
  public static <Match extends IPatternMatch> ViatraQueryTest test(final IQuerySpecification<? extends ViatraQueryMatcher<Match>> pattern) {
    ViatraQueryTest _viatraQueryTest = new ViatraQueryTest();
    return _viatraQueryTest.and(pattern);
  }
  
  public static ViatraQueryTest test(final IQueryGroup patterns) {
    ViatraQueryTest _viatraQueryTest = new ViatraQueryTest();
    return _viatraQueryTest.and(patterns);
  }
  
  public static ViatraQueryTest test() {
    return new ViatraQueryTest();
  }
  
  /**
   * Test the specified query
   */
  public static <Match extends IPatternMatch> ViatraQueryTest test(final String pattern) {
    ViatraQueryTest _test = ViatraQueryTest.test();
    return _test.and(pattern);
  }
  
  public ViatraQueryTest and(final IQueryGroup patterns) {
    ViatraQueryTest _xblockexpression = null;
    {
      Set<IQuerySpecification<?>> _specifications = patterns.getSpecifications();
      final Procedure1<IQuerySpecification<?>> _function = new Procedure1<IQuerySpecification<?>>() {
        @Override
        public void apply(final IQuerySpecification<?> it) {
          ViatraQueryTest.this.patterns.add(((IQuerySpecification<ViatraQueryMatcher<IPatternMatch>>) it));
        }
      };
      IterableExtensions.<IQuerySpecification<?>>forEach(_specifications, _function);
      _xblockexpression = this;
    }
    return _xblockexpression;
  }
  
  public ViatraQueryTest and(final IQuerySpecification<? extends ViatraQueryMatcher<? extends IPatternMatch>> pattern) {
    ViatraQueryTest _xblockexpression = null;
    {
      this.patterns.add(pattern);
      _xblockexpression = this;
    }
    return _xblockexpression;
  }
  
  public ViatraQueryTest and(final String pattern) {
    IQuerySpecification<? extends ViatraQueryMatcher<? extends IPatternMatch>> _querySpecification = QuerySpecificationRegistry.getQuerySpecification(pattern);
    return this.and(
      ((IQuerySpecification<ViatraQueryMatcher<IPatternMatch>>) _querySpecification));
  }
  
  private final List<IQuerySpecification<? extends ViatraQueryMatcher<? extends IPatternMatch>>> patterns = new LinkedList<IQuerySpecification<? extends ViatraQueryMatcher<? extends IPatternMatch>>>();
  
  private ViatraQueryTest() {
    ViatraQueryTestCase _viatraQueryTestCase = new ViatraQueryTestCase();
    this.testCase = _viatraQueryTestCase;
  }
  
  private ViatraQueryTest(final IQuerySpecification<? extends ViatraQueryMatcher<? extends IPatternMatch>> pattern) {
    this();
    this.patterns.add(pattern);
  }
  
  /**
   * Add match result set with a query initialized using the given hints
   */
  public ViatraQueryTest with(final QueryEvaluationHint hint) {
    ViatraQueryTest _xblockexpression = null;
    {
      final PatternBasedMatchSetModelProvider modelProvider = new PatternBasedMatchSetModelProvider(hint);
      this.testCase.addMatchSetModelProvider(modelProvider);
      _xblockexpression = this;
    }
    return _xblockexpression;
  }
  
  /**
   * Add match result set with a query initialized using the given query backend
   */
  public ViatraQueryTest with(final Class<? extends IQueryBackend> queryBackendClass) {
    ViatraQueryTest _xblockexpression = null;
    {
      Map<String, Object> _emptyMap = CollectionLiterals.<String, Object>emptyMap();
      final QueryEvaluationHint hint = new QueryEvaluationHint(queryBackendClass, _emptyMap);
      _xblockexpression = this.with(hint);
    }
    return _xblockexpression;
  }
  
  /**
   * Add match result set loaded from the given snapshot
   */
  public ViatraQueryTest with(final URI snapshotURI) {
    ViatraQueryTest _xblockexpression = null;
    {
      SnapshotMatchSetModelProvider _snapshotMatchSetModelProvider = new SnapshotMatchSetModelProvider(snapshotURI);
      this.testCase.addMatchSetModelProvider(_snapshotMatchSetModelProvider);
      _xblockexpression = this;
    }
    return _xblockexpression;
  }
  
  /**
   * Add match result set loaded from the given snapshot
   */
  public ViatraQueryTest with(final String snapshotURI) {
    URI _resolvePlatformURI = XmiModelUtil.resolvePlatformURI(XmiModelUtil.XmiModelUtilRunningOptionEnum.BOTH, snapshotURI);
    return this.with(_resolvePlatformURI);
  }
  
  /**
   * Load input model
   */
  public ViatraQueryTest on(final URI inputURI) {
    ViatraQueryTest _xblockexpression = null;
    {
      this.testCase.loadModel(inputURI);
      _xblockexpression = this;
    }
    return _xblockexpression;
  }
  
  /**
   * Execute all queries and check that the result sets are equal
   */
  public void assertEquals() {
    final Procedure1<IQuerySpecification<? extends ViatraQueryMatcher<? extends IPatternMatch>>> _function = new Procedure1<IQuerySpecification<? extends ViatraQueryMatcher<? extends IPatternMatch>>>() {
      @Override
      public void apply(final IQuerySpecification<? extends ViatraQueryMatcher<? extends IPatternMatch>> it) {
        ViatraQueryTest.this.testCase.<IPatternMatch>assertMatchSetsEqual(((IQuerySpecification<ViatraQueryMatcher<IPatternMatch>>) it));
      }
    };
    IterableExtensions.<IQuerySpecification<? extends ViatraQueryMatcher<? extends IPatternMatch>>>forEach(this.patterns, _function);
  }
}
